/*
 * Copyright (C) 2021, Xilinx Inc - All rights reserved
 *
 * Licensed under the Apache License, Version 2.0 (the "License"). You may
 * not use this file except in compliance with the License. A copy of the
 * License is located at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations
 * under the License.
 */

#pragma once

#include <mutex>
#include <condition_variable>
#include <xmacpp/Log.h>
#include "Exception.h"
#include "SessionBase.h"
#include "InternalFrame.h"

namespace xma {

//////////////////////////////////////////////////////////////////////////////
class EmptyDynamicParams {
};

template <class DynParams = EmptyDynamicParams>
class FilterThread;

template <class DynParams = EmptyDynamicParams>
struct ThreadInfo {
  FilterThread<DynParams>* parent;
  std::thread              thread;
  std::condition_variable  enqueue_cond;
  queue_t                  enqueue_data;
  DynParams                dynamic_params;
  bool                     has_enqueue_data = false;
  std::condition_variable  dequeue_cond;
  queue_t                  dequeue_data;
  bool                     has_dequeue_data = false;
  bool                     is_processing    = false;
  uint32_t                 extend           = 0;
  ExceptionStore           exception;
  int32_t                  ret = XMA_SUCCESS;
};

#define XMA_FILTER_THREAD "FilterThread"

//////////////////////////////////////////////////////////////////////////////
void MoveQueue(queue_t& dst, queue_t& src);

//////////////////////////////////////////////////////////////////////////////
template <class DynParams>
class FilterThread : public SessionBase {
public:
  FilterThread(IntSessionFactoryPtr sf, XmaParameter* params, uint32_t param_cnt, uint32_t default_threads, uint32_t max_threads) : SessionBase(sf) {
    m_log->LogMsg(XMA_DEBUG_LOG, XMA_FILTER_THREAD, "FilterThread::FilterThread()");

    m_num_threads       = default_threads;
    XmaParameter* param = GetParameter(params, param_cnt, XMA_FILTER_PARAM_THREADS);
    if (param) {
      m_num_threads = *static_cast<uint32_t*>(param->value);
    }
    if (m_num_threads > max_threads) {
      m_num_threads = max_threads;
    }

    if (m_num_threads == 0) {
      m_log->LogMsg(XMA_DEBUG_LOG, XMA_FILTER_THREAD, "FilterThread::FilterThread() - Pipeline disabled");

      thread_info_t thread_info = std::make_unique<ThreadInfo<DynParams>>();
      m_threads.push_back(std::move(thread_info));
    } else {
      m_log->LogMsg(XMA_DEBUG_LOG, XMA_FILTER_THREAD, "FilterThread::FilterThread() - Pipelining %d threads", m_num_threads);

      uint32_t thread;
      for (thread = 0; thread < m_num_threads; ++thread) {
        thread_info_t thread_info = std::make_unique<ThreadInfo<DynParams>>();
        thread_info->parent       = this;
        thread_info->thread       = std::thread(FilterThread::Thread, thread_info.get());
        m_threads.push_back(std::move(thread_info));
      }
      m_threads[0]->extend = (2 * m_num_threads) - 1;

      param = GetParameter(params, param_cnt, XMA_FILTER_PARAM_WAIT);
      if (param) {
        m_wait = *static_cast<int32_t*>(param->value);
      }
    }

    m_log->LogMsg(XMA_DEBUG_LOG, XMA_FILTER_THREAD, "FilterThread::FilterThread() - returning");
  }

  virtual ~FilterThread() { m_log->LogMsg(XMA_DEBUG_LOG, XMA_FILTER_THREAD, "FilterThread::~FilterThread()"); }

  int32_t SetLog(LogPtr log) override {
    lock_t lock(m_mutex);
    SessionBase::SetLog(log);
    return XMA_SUCCESS;
  }

  int32_t Push(queue_t& data, DynParams* params = nullptr) {
    m_log->LogMsg(XMA_DEBUG_LOG, XMA_FILTER_THREAD, "FilterThread::Push()");
    lock_t lock(m_mutex);
    if (m_threads[m_next]->has_enqueue_data) {
      m_log->LogMsg(XMA_DEBUG_LOG, XMA_FILTER_THREAD, "FilterThread::Push() - returning try again");
      return XMA_TRY_AGAIN;
    }

    switch (data.index()) {
    case 0: // QueueFrame
      if (!std::get<QueueFrame>(data).frame) {
        m_eos_pushed = true;
      }
      break;
    case 1: // QueueFrameList
      if (std::get<QueueFrameList>(data).frame_list.empty()) {
        m_eos_pushed = true;
      }
      break;
    case 2: //QueueDataBuffer
      if (!std::get<QueueDataBuffer>(data).data) {
        m_eos_pushed = true;
      }
      break;
    }

    m_threads[m_next]->has_enqueue_data = true;
    MoveQueue(m_threads[m_next]->enqueue_data, data);
    if (params) {
      memcpy(&m_threads[m_next]->dynamic_params, params, sizeof(DynParams));
    }
    if (m_num_threads > 0) {
      m_threads[m_next]->enqueue_cond.notify_one();
      m_next = (m_next + 1) % m_num_threads;
      ++m_processing;
    }

    m_log->LogMsg(XMA_DEBUG_LOG, XMA_FILTER_THREAD, "FilterThread::Push() - returning success");
    return XMA_SUCCESS;
  }

  int32_t Pull(queue_t& data) {
    m_log->LogMsg(XMA_DEBUG_LOG, XMA_FILTER_THREAD, "FilterThread::Pull()");

    lock_t lock(m_mutex);

    if (m_wait) {
      if (m_processing == 0) {
        m_log->LogMsg(XMA_DEBUG_LOG, XMA_FILTER_THREAD, "FilterThread::Pull() - returning try again");
        return XMA_TRY_AGAIN;
      }
      while (!m_threads[m_last]->has_dequeue_data) {
        m_wait_cond.wait(lock);
      }
    }

    if (m_num_threads > 0) {
      if (!m_threads[m_last]->has_dequeue_data) {
        if (m_processing == 0) {
          m_log->LogMsg(XMA_DEBUG_LOG, XMA_FILTER_THREAD, "FilterThread::Pull() - returning resend and recv");
          return XMA_RESEND_AND_RECV;
        }
        if (m_threads[m_last]->is_processing || m_threads[m_last]->has_enqueue_data) {
          m_log->LogMsg(XMA_DEBUG_LOG, XMA_FILTER_THREAD, "FilterThread::Pull() - returning try again");
          return XMA_TRY_AGAIN;
        } else {
          m_log->LogMsg(XMA_DEBUG_LOG, XMA_FILTER_THREAD, "FilterThread::Pull() - returning resend and recv");
          return XMA_RESEND_AND_RECV;
        }
      }
      if (m_threads[m_last]->exception.HasException()) {
        m_threads[m_last]->exception.RethrowException();
      }
      m_threads[m_last]->has_dequeue_data = false;
      MoveQueue(data, m_threads[m_last]->dequeue_data);
      m_threads[m_last]->dequeue_cond.notify_one();
      int32_t ret = m_threads[m_last]->ret;
      m_last      = (m_last + 1) % m_num_threads;
      --m_processing;

      m_log->LogMsg(XMA_DEBUG_LOG, XMA_FILTER_THREAD, "FilterThread::Pull() - returning %d", ret);
      return ret;
    } else {
      if (!m_threads[0]->has_enqueue_data) {
        m_log->LogMsg(XMA_DEBUG_LOG, XMA_FILTER_THREAD, "FilterThread::Pull() - returning resend and recv");
        return XMA_RESEND_AND_RECV;
      }

      int32_t ret = Process(m_threads[0]->enqueue_data, data, &m_threads[0]->dynamic_params);

      m_threads[0]->enqueue_data     = QueueFrame();
      m_threads[0]->has_enqueue_data = false;

      m_log->LogMsg(XMA_DEBUG_LOG, XMA_FILTER_THREAD, "FilterThread::Pull() - returning %d", ret);
      return ret;
    }
  }

protected:
  virtual int32_t Process(queue_t& in, queue_t& out, DynParams* dynamic_params) = 0;

  void Close() {
    m_log->LogMsg(XMA_DEBUG_LOG, XMA_FILTER_THREAD, "FilterThread::Close()");

    lock_t lock(m_mutex);
    if (!m_eos_pushed) {
      m_log->LogMsg(XMA_WARNING_LOG, XMA_FILTER_THREAD, "FilterThread::Close() - End of stream not sent!");
    } else if (m_num_threads > 0) {
      if (m_processing > 0) {
        m_log->LogMsg(XMA_WARNING_LOG, XMA_FILTER_THREAD, "FilterThread::Close() - End of stream not recieved!");
      }
    } else if (m_threads[0]->has_enqueue_data) {
      m_log->LogMsg(XMA_WARNING_LOG, XMA_FILTER_THREAD, "FilterThread::Close() - End of stream not recieved!");
    }

    m_done = true;

    if (m_num_threads > 0) {
      uint32_t thread;
      for (thread = 0; thread < m_num_threads; ++thread) {
        if (m_threads[thread]->has_enqueue_data) {
          m_threads[thread]->has_enqueue_data = false;
          m_threads[thread]->enqueue_data     = QueueFrame();
        }
        m_threads[thread]->enqueue_cond.notify_one();
        if (m_threads[thread]->has_dequeue_data) {
          m_threads[thread]->has_dequeue_data = false;
          m_threads[thread]->dequeue_data     = QueueFrame();
        }
        m_threads[thread]->dequeue_cond.notify_one();
      }

      lock.unlock();

      for (thread = 0; thread < m_num_threads; ++thread) {
        m_threads[thread]->thread.join();
      }
    } else {
      m_threads[0]->has_enqueue_data = false;
      m_threads[0]->enqueue_data     = QueueFrame();
      m_threads[0]->has_dequeue_data = false;
      m_threads[0]->dequeue_data     = QueueFrame();
    }

    m_log->LogMsg(XMA_DEBUG_LOG, XMA_FILTER_THREAD, "FilterThread::Close() - returning");
  }

private:
  static void Thread(ThreadInfo<DynParams>* thread) {
    lock_t lock(thread->parent->m_mutex);
    while (!thread->parent->m_done) {
      while (!thread->has_enqueue_data) {
        thread->enqueue_cond.wait(lock);
        if (thread->parent->m_done) {
          return;
        }
      }
      queue_t in;
      MoveQueue(in, thread->enqueue_data);
      DynParams params;
      memcpy(&params, &thread->dynamic_params, sizeof(DynParams));
      thread->has_enqueue_data = false;
      thread->is_processing    = true;
      lock.unlock();

      queue_t out;

      int32_t ret = XMA_SUCCESS;
      try {
        if (thread->extend > 0) {
          ExtendBufferPools(in, thread->extend);
        }
        ret = thread->parent->Process(in, out, &params);
        if (thread->extend > 0) {
          ExtendBufferPools(out, thread->extend);
          thread->extend = 0;
        }

        lock.lock();
        while (thread->has_dequeue_data) {
          if (thread->parent->m_done) {
            return;
          }
          thread->dequeue_cond.wait(lock);
        }
      } catch (ThreadThrowable& exception) {
        lock.lock();
        while (thread->has_dequeue_data) {
          if (thread->parent->m_done) {
            return;
          }
          thread->dequeue_cond.wait(lock);
        }
        thread->exception.SetException(exception.CreateClone());
      } catch (...) {
        lock.lock();
        while (thread->has_dequeue_data) {
          if (thread->parent->m_done) {
            return;
          }
          thread->dequeue_cond.wait(lock);
        }
        thread->exception.SetException(std::make_shared<UnknownException>());
      }

      thread->is_processing = false;
      MoveQueue(thread->dequeue_data, out);
      thread->has_dequeue_data = true;
      if (thread->parent->m_wait) {
        thread->parent->m_wait_cond.notify_one();
      }
      thread->ret = ret;
    }
  }

  static void ExtendBufferPools(queue_t& data, uint32_t& frames) {
    switch (data.index()) {
    case 0: { // QueueFrame
      IntFramePtr frame = std::dynamic_pointer_cast<InternalFrame>(std::get<QueueFrame>(data).frame);
      if (frame) {
        frame->ExtendBufferPool(frames);
      }
      break;
    }
    case 1: { // QueueFrameList
      std::vector<FramePtr> frame_list = std::get<QueueFrameList>(data).frame_list;
      size_t                size       = frame_list.size();
      size_t                idx;
      for (idx = 0; idx < size; ++idx) {
        IntFramePtr frame = std::dynamic_pointer_cast<InternalFrame>(frame_list[idx]);
        if (frame) {
          frame->ExtendBufferPool(frames);
        }
      }
      break;
    }
    case 2: //QueueDataBuffer
      break;
    }
  }

protected:
  using lock_t        = std::unique_lock<std::mutex>;
  using thread_info_t = std::unique_ptr<ThreadInfo<DynParams>>;
  std::mutex                 m_mutex;
  uint32_t                   m_num_threads = 1;
  int32_t                    m_wait        = 0;
  std::condition_variable    m_wait_cond;
  std::vector<thread_info_t> m_threads;
  uint32_t                   m_next       = 0;
  uint32_t                   m_last       = 0;
  uint32_t                   m_processing = 0;
  bool                       m_eos_pushed = false;
  bool                       m_done       = false;
};

}; // Namespace xma
