/*
 * Copyright (C) 2021, Xilinx Inc - All rights reserved
 *
 * Licensed under the Apache License, Version 2.0 (the "License"). You may
 * not use this file except in compliance with the License. A copy of the
 * License is located at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations
 * under the License.
 */

#pragma once

#include <deque>
#include <mutex>
#include <functional>
#include "xmacpp/ScalerSession.h"
#include "SessionBase.h"
#include "FilterThread.h"

namespace xma {
//////////////////////////////////////////////////////////////////////////////
class ScalerDynamicParams {
public:
  int32_t left;
  int32_t top;
  int32_t width;
  int32_t height;
};

//////////////////////////////////////////////////////////////////////////////
class InternalScalerSession final : virtual public ScalerSession, virtual public FilterThread<ScalerDynamicParams> {
public:
  InternalScalerSession(IntSessionFactoryPtr sf, XmaScalerProperties* props);
  virtual ~InternalScalerSession();

  virtual int32_t Enqueue(queue_t& data) override;
  virtual int32_t Dequeue(queue_t& data) override;

private:
  void                                   InitScaler(int32_t left, int32_t top, int32_t width, int32_t height);
  std::shared_ptr<InternalScalerSession> GetMixrateScaler(void* mixrate, XmaScalerProperties* props);
  void                                   SetMixrateChild(InternalScalerSession* child);
  void                                   SendMixrateEOS();
  void                                   RecvMixrateEOS();

  virtual int32_t Process(queue_t& in, queue_t& out, ScalerDynamicParams* dynamic_params) override;

protected:
  using vpi_mem_handle_t    = std::unique_ptr<void, std::function<void(VpiMemHandle mem_ctx)>>;
  using scale_handle_t      = std::unique_ptr<void, std::function<void(void*)>>;
  using frame_props_array_t = std::vector<FramePropertiesPtr>;
  using vpi_frms_t          = std::unique_ptr<VpiFrm*, std::function<void(VpiFrm**)>>;
  vpi_mem_handle_t                       m_mem_ctx;
  VpiProcInitParams                      m_params;
  scale_handle_t                         m_handle;
  frame_props_array_t                    m_props;
  int32_t                                m_outputs = 1;
  XmaFraction                            m_framerate;
  std::shared_ptr<InternalScalerSession> m_mixrate_parent_scaler;
  InternalScalerSession*                 m_mixrate_scaler   = nullptr;
  char                                   m_resize_str[1024] = {'\0'};
  int32_t                                m_dynamic_cropping = 0;
  int32_t                                m_left             = 0;
  int32_t                                m_top              = 0;
  int32_t                                m_width            = 0;
  int32_t                                m_height           = 0;
  int32_t*                               m_dynamic_left     = nullptr;
  int32_t*                               m_dynamic_top      = nullptr;
  int32_t*                               m_dynamic_width    = nullptr;
  int32_t*                               m_dynamic_height   = nullptr;
  int32_t                                m_old_left         = 0;
  int32_t                                m_old_top          = 0;
  int32_t                                m_old_width        = 0;
  int32_t                                m_old_height       = 0;
  int64_t                                m_frame_sent       = 0;
  int64_t                                m_frame_recv       = 0;
  bool                                   m_latency_logging  = false;
  struct timespec                        m_latency;
  int64_t                                m_time_taken;
};

using IntScalerSessionPtr = std::shared_ptr<InternalScalerSession>;

}; // Namespace xma
