/*
 * Copyright (C) 2021, Xilinx Inc - All rights reserved
 *
 * Licensed under the Apache License, Version 2.0 (the "License"). You may
 * not use this file except in compliance with the License. A copy of the
 * License is located at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations
 * under the License.
 */

#pragma once

#include "xma.h"
#include "xmacpp/Log.h"
#include <list>
#include <atomic>
#include <mutex>
#include <map>
#include <list>
#include <syslog.h>
#include <fstream>

namespace xma {

class XmaLog : public Log {
public:
  XmaLog(XmaLogLevelType level) : m_level(level) {
  }
  virtual ~XmaLog() {
  }

  virtual void LogMsgList(XmaLogLevelType level, const char* name, const char* msg, va_list arg) override;

  virtual XmaLogLevelType GetLogLevel() const override;
  virtual const char*     GetFileName() const override;

protected:
  virtual void Log(XmaLogLevelType level, const char* tag, const char* msg) = 0;

protected:
  XmaLogLevelType m_level = XMA_ERROR_LOG;
  std::mutex      m_mutex;
};

using XmaLogPtr = std::shared_ptr<XmaLog>;

class XmaLogNull final : public XmaLog {
public:
  explicit XmaLogNull(XmaLogLevelType level) : XmaLog(level) {
  }
  virtual ~XmaLogNull() {
  }

  virtual XmaLogType GetLogType() const override { return XMA_LOG_TYPE_NONE; }

protected:
  virtual void Log(XmaLogLevelType, const char*, const char*) override {
  }
};

class XmaLogConsole final : public XmaLog {
public:
  explicit XmaLogConsole(XmaLogLevelType level) : XmaLog(level) {
  }
  virtual ~XmaLogConsole() {
  }

  virtual XmaLogType GetLogType() const override { return XMA_LOG_TYPE_CONSOLE; }

protected:
  virtual void Log(XmaLogLevelType level, const char* tag, const char* msg) override;

private:
  std::map<XmaLogLevelType, const char*> severityMap = {{XmaLogLevelType::XMA_EMERGENCY_LOG, "EMERGENCY: "}, {XmaLogLevelType::XMA_ALERT_LOG, "ALERT: "},
      {XmaLogLevelType::XMA_CRITICAL_LOG, "CRITICAL: "}, {XmaLogLevelType::XMA_ERROR_LOG, "ERROR: "}, {XmaLogLevelType::XMA_WARNING_LOG, "WARNING: "},
      {XmaLogLevelType::XMA_NOTICE_LOG, "NOTICE: "}, {XmaLogLevelType::XMA_INFO_LOG, "INFO: "}, {XmaLogLevelType::XMA_DEBUG_LOG, "DEBUG: "}};
};

class XmaLogSyslog final : public XmaLog {
public:
  explicit XmaLogSyslog(XmaLogLevelType level) : XmaLog(level) { openlog(NULL, LOG_PID | LOG_CONS, LOG_USER); }
  virtual ~XmaLogSyslog() { closelog(); }

  virtual XmaLogType GetLogType() const override { return XMA_LOG_TYPE_SYSLOG; }

protected:
  virtual void Log(XmaLogLevelType level, const char* tag, const char* msg) override;

private:
  std::map<XmaLogLevelType, int> severityMap = {{XmaLogLevelType::XMA_EMERGENCY_LOG, LOG_EMERG}, {XmaLogLevelType::XMA_ALERT_LOG, LOG_ALERT},
      {XmaLogLevelType::XMA_CRITICAL_LOG, LOG_CRIT}, {XmaLogLevelType::XMA_ERROR_LOG, LOG_ERR}, {XmaLogLevelType::XMA_WARNING_LOG, LOG_WARNING},
      {XmaLogLevelType::XMA_NOTICE_LOG, LOG_NOTICE}, {XmaLogLevelType::XMA_INFO_LOG, LOG_INFO}, {XmaLogLevelType::XMA_DEBUG_LOG, LOG_DEBUG}};
};

class XmaLogFile final : public XmaLog {
public:
  XmaLogFile(XmaLogLevelType level, const char* filename);
  virtual ~XmaLogFile();

  virtual XmaLogType  GetLogType() const override { return XMA_LOG_TYPE_FILE; }
  virtual const char* GetFileName() const override { return m_filename; }

protected:
  virtual void Log(XmaLogLevelType level, const char* tag, const char* msg) override;

private:
  const char*                            m_filename;
  std::ofstream                          m_handle;
  std::map<XmaLogLevelType, const char*> severityMap = {{XmaLogLevelType::XMA_EMERGENCY_LOG, "EMERGENCY: "}, {XmaLogLevelType::XMA_ALERT_LOG, "ALERT: "},
      {XmaLogLevelType::XMA_CRITICAL_LOG, "CRITICAL: "}, {XmaLogLevelType::XMA_ERROR_LOG, "ERROR: "}, {XmaLogLevelType::XMA_WARNING_LOG, "WARNING: "},
      {XmaLogLevelType::XMA_NOTICE_LOG, "NOTICE: "}, {XmaLogLevelType::XMA_INFO_LOG, "INFO: "}, {XmaLogLevelType::XMA_DEBUG_LOG, "DEBUG: "}};
};

class XmaLogCallback final : public XmaLog {
public:
  XmaLogCallback(XmaLogLevelType level, XmaLogCallbackFunction callback, void* opaque) : XmaLog(level), m_callback(callback), m_opaque(opaque) {
  }
  virtual ~XmaLogCallback() {
  }

  virtual void LogMsgList(XmaLogLevelType level, const char* name, const char* msg, va_list arg) override;

  virtual XmaLogType GetLogType() const override { return XMA_LOG_TYPE_CALLBACK; }

protected:
  virtual void Log(XmaLogLevelType level, const char* tag, const char* msg) override;

private:
  XmaLogCallbackFunction m_callback;
  void*                  m_opaque;
};

} // namespace xma
