/*
 * Copyright (C) 2021, Xilinx Inc - All rights reserved
 *
 * Licensed under the Apache License, Version 2.0 (the "License"). You may
 * not use this file except in compliance with the License. A copy of the
 * License is located at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations
 * under the License.
 */

#include <cstring>
#include "xmaapidecoder.h"
#include "xmaapi.h"
#include "InternalFrame.h"
#include "InternalDataBuffer.h"

#define XMA_DECODER_MOD "xmadecoder"

///////////////////////////////////////////////////////////////////////////////
XmaDecoderSession* xma_dec_session_create(XmaDecoderProperties* dec_props) {
  if (!dec_props) {
    return nullptr;
  }

  XmaSessionFactory* xma_sf = static_cast<XmaSessionFactory*>(dec_props->handle);
  if (!xma_sf) {
    return nullptr;
  }

  TRY()
  xma_sf->m_sf->GetLog()->LogMsg(XMA_DEBUG_LOG, XMA_DECODER_MOD, "xma_dec_session_create()");
  XmaDecoderSession* session = new XmaDecoderSession();
  if (!session) {
    xma_sf->m_sf->GetLog()->LogMsg(XMA_ERROR_LOG, XMA_DECODER_MOD, "xma_dec_session_create() Out Of Memory!");
    return nullptr;
  }

  session->m_sf  = xma_sf->m_sf;
  session->m_log = xma_sf->m_sf->GetLog();

  session->m_session = xma_sf->m_sf->MakeDecoderSession(dec_props);
  if (!session->m_session) {
    session->m_log->LogMsg(XMA_ERROR_LOG, XMA_DECODER_MOD, "xma_dec_session_create() Out Of Memory!");
    delete session;
    return nullptr;
  }

  xma_sf->m_sf->GetLog()->LogMsg(XMA_DEBUG_LOG, XMA_DECODER_MOD, "xma_dec_session_create() - Returning");
  return static_cast<XmaDecoderSession*>(session);
  CATCH_NULL(xma_sf->m_sf->GetLog(), XMA_DECODER_MOD)
}

///////////////////////////////////////////////////////////////////////////////
int32_t xma_dec_session_destroy(XmaDecoderSession* session) {
  if (!session) {
    return XMA_ERROR_INVALID;
  }

  xma::LogPtr log = session->m_log;
  log->LogMsg(XMA_DEBUG_LOG, XMA_DECODER_MOD, "xma_dec_session_destroy()");

  if (!session->m_session) {
    log->LogMsg(XMA_ERROR_LOG, XMA_DECODER_MOD, "xma_dec_session_destroy() Invalid parameter!");
    delete session;
    return XMA_ERROR_INVALID;
  }

  TRY()
  delete session;

  log->LogMsg(XMA_DEBUG_LOG, XMA_DECODER_MOD, "xma_dec_session_destroy() - Returning success");
  return XMA_SUCCESS;
  CATCH_ERROR(log, XMA_DECODER_MOD)
}

///////////////////////////////////////////////////////////////////////////////
int32_t xma_dec_session_set_log(XmaDecoderSession* session, XmaLogHandle handle) {
  if (!session || !handle) {
    return XMA_ERROR_INVALID;
  }

  TRY()
  XmaLogStruct* log_struct = static_cast<XmaLogStruct*>(handle);
  xma::LogPtr   log        = log_struct->m_log;
  const char*   level;
  switch (log->GetLogLevel()) {
  case XMA_EMERGENCY_LOG:
    level = "emergency";
    break;
  case XMA_ALERT_LOG:
    level = "alert";
    break;
  case XMA_CRITICAL_LOG:
    level = "critical";
    break;
  case XMA_ERROR_LOG:
    level = "error";
    break;
  case XMA_WARNING_LOG:
    level = "warning";
    break;
  case XMA_NOTICE_LOG:
    level = "notice";
    break;
  case XMA_INFO_LOG:
    level = "info";
    break;
  case XMA_DEBUG_LOG:
    level = "debug";
    break;
  default:
    level = "unknown";
    break;
  }
  switch (log->GetLogType()) {
  case XMA_LOG_TYPE_NONE:
    session->m_log->LogMsg(XMA_DEBUG_LOG, XMA_DECODER_MOD, "xma_dec_session_set_log(level=%s, type=none)", level);
    break;
  case XMA_LOG_TYPE_CONSOLE:
    session->m_log->LogMsg(XMA_DEBUG_LOG, XMA_DECODER_MOD, "xma_dec_session_set_log(level=%s, type=console)", level);
    break;
  case XMA_LOG_TYPE_SYSLOG:
    session->m_log->LogMsg(XMA_DEBUG_LOG, XMA_DECODER_MOD, "xma_dec_session_set_log(level=%s, type=syslog)", level);
    break;
  case XMA_LOG_TYPE_FILE:
    session->m_log->LogMsg(XMA_DEBUG_LOG, XMA_DECODER_MOD, "xma_dec_session_set_log(level=%s, type=file, file='%s')", level, log->GetFileName());
    break;
  case XMA_LOG_TYPE_CALLBACK:
    session->m_log->LogMsg(XMA_DEBUG_LOG, XMA_DECODER_MOD, "xma_dec_session_set_log(level=%s, type=callback)", level);
    break;
  default:
    session->m_log->LogMsg(XMA_DEBUG_LOG, XMA_DECODER_MOD, "xma_dec_session_set_log(level=%s, type=unknown)", level);
    break;
  }
  session->m_log = log;
  session->m_session->SetLog(log);

  session->m_log->LogMsg(XMA_DEBUG_LOG, XMA_DECODER_MOD, "xma_dec_session_set_log() - Returning success!");
  return XMA_SUCCESS;
  CATCH_ERROR(session->m_log, XMA_DECODER_MOD)
}

///////////////////////////////////////////////////////////////////////////////
int32_t xma_dec_session_send_data(XmaDecoderSession* session, XmaDataBuffer* data, int32_t* data_used) {
  if (!session) {
    return XMA_ERROR_INVALID;
  }

  if (!session->m_session) {
    session->m_log->LogMsg(XMA_ERROR_LOG, XMA_DECODER_MOD, "xma_dec_session_send_data() Invalid parameter!");
    return XMA_ERROR_INVALID;
  }

  TRY()
  session->m_log->LogMsg(XMA_DEBUG_LOG, XMA_DECODER_MOD, "xma_dec_session_send_data()");

  xma::queue_t queue = xma::QueueDataBuffer();
  if (data) {
    xma::PrivateDataBufferInfo* info = static_cast<xma::PrivateDataBufferInfo*>(data->private_do_not_touch);
    xma::IntDataBufferPtr       buf;
    if ((info) && (info->IsValid())) {
      buf = std::dynamic_pointer_cast<xma::InternalDataBuffer>(info->GetSelf());
    }
    if (buf) {
      std::get<xma::QueueDataBuffer>(queue).data = buf;
      if (buf->GetRefCount() == 1) {
        info->SetSelf(nullptr);
      } else {
        buf->DecRef();
      }
    }

    if (data_used) {
      *data_used = data->alloc_size;
    }
  }

  int32_t ret = session->m_session->Enqueue(queue);
  char    error_msg[256];
  session->m_log->LogMsg(XMA_DEBUG_LOG, XMA_DECODER_MOD, "xma_dec_session_send_data() - Returning %s", xma_perror(ret, error_msg, sizeof(error_msg)));
  return ret;
  CATCH_ERROR(session->m_log, XMA_DECODER_MOD)
}

///////////////////////////////////////////////////////////////////////////////
int32_t xma_dec_session_get_properties(XmaDecoderSession* session, XmaFrameProperties* fprops) {
  if (!session) {
    return XMA_ERROR_INVALID;
  }

  if (!session->m_session || !fprops) {
    session->m_log->LogMsg(XMA_ERROR_LOG, XMA_DECODER_MOD, "xma_dec_session_get_properties() Invalid parameter!");
    return XMA_ERROR_INVALID;
  }

  TRY()
  session->m_log->LogMsg(XMA_DEBUG_LOG, XMA_DECODER_MOD, "xma_dec_session_get_properties()");
  xma::FramePropertiesPtr props = session->m_session->GetProps();
  if (props) {
    fprops->format    = props->GetFormat();
    fprops->sw_format = props->GetSwFormat();
    fprops->width     = props->GetWidth();
    fprops->height    = props->GetHeight();
    int32_t planes    = props->GetNumPlanes();
    for (int32_t plane = 0; plane < planes; ++plane) {
      fprops->linesize[plane] = props->GetPlaneStride(plane);
    }

    session->m_log->LogMsg(XMA_DEBUG_LOG, XMA_DECODER_MOD, "xma_dec_session_get_properties() - Returning success");
    return XMA_SUCCESS;
  } else {
    session->m_log->LogMsg(XMA_DEBUG_LOG, XMA_DECODER_MOD, "xma_dec_session_get_properties() - Returning send more data");
    return XMA_SEND_MORE_DATA;
  }
  CATCH_ERROR(session->m_log, XMA_DECODER_MOD)
}

///////////////////////////////////////////////////////////////////////////////
int32_t xma_dec_session_recv_frame(XmaDecoderSession* session, XmaFrame* xma_frame) {
  if (!session) {
    return XMA_ERROR_INVALID;
  }

  if (!session->m_session) {
    session->m_log->LogMsg(XMA_ERROR_LOG, XMA_DECODER_MOD, "xma_dec_session_recv_frame() Invalid parameter!");
    return XMA_ERROR_INVALID;
  }

  if (!xma_frame || (xma_frame->data[0].buffer_type != NO_BUFFER) || (xma_frame->data[0].refcount != 1)) {
    session->m_log->LogMsg(XMA_ERROR_LOG, XMA_DECODER_MOD, "xma_dec_session_recv_frame() Invalid parameter!");
    return XMA_ERROR_INVALID;
  }

  TRY()
  session->m_log->LogMsg(XMA_DEBUG_LOG, XMA_DECODER_MOD, "xma_dec_session_recv_frame()");
  xma::queue_t queue;
  int32_t      ret = session->m_session->Dequeue(queue);
  if (ret != XMA_SUCCESS) {
    char error_msg[256];
    session->m_log->LogMsg(XMA_DEBUG_LOG, XMA_DECODER_MOD, "xma_dec_session_recv_frame() - Returning %s", xma_perror(ret, error_msg, sizeof(error_msg)));
    return ret;
  }

  CopyFrame(session->m_log, xma_frame, std::move(std::get<xma::QueueFrame>(queue).frame));

  session->m_log->LogMsg(XMA_DEBUG_LOG, XMA_DECODER_MOD, "xma_dec_session_recv_frame() - Returning success!");
  return ret;
  CATCH_ERROR(session->m_log, XMA_DECODER_MOD)
}
